from __future__ import print_function, division, absolute_import#, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

import distributed
from dask import array, delayed
from rfpipe import source, search, util, candidates
from dask.base import tokenize
import numpy as np

import logging
logger = logging.getLogger(__name__)
vys_timeout_default = 10


def pipeline_scan(st, segments=None, cl=None, host=None, cfile=None,
                  vys_timeout=vys_timeout_default):
    """ Given rfpipe state and dask distributed client, run search pipline.
    throttle option will submit only if worker state allows for it.
    parameters of throttling:
      read_overhead scales vismem to peak usage during read.
      read_totfrac is total memory usage allowed in bytes at submission time.
    """

    if cl is None:
        if host is None:
            cl = distributed.Client(n_workers=1, threads_per_worker=16,
                                    resources={"READER": 1, "CORES": 16},
                                    local_dir="/lustre/evla/test/realfast/scratch")
        else:
            cl = distributed.Client('{0}:{1}'.format(host, '8786'))

    if not isinstance(segments, list):
        segments = list(range(st.nsegment))

    futures = []
    for segment in segments:
        futures.append(pipeline_seg(st, segment, cl=cl, cfile=cfile,
                                    vys_timeout=vys_timeout))

    return futures  # list of dicts


def pipeline_seg(st, segment, cl, cfile=None,
                 vys_timeout=vys_timeout_default):
    """ Submit pipeline processing of a single segment to scheduler.
    Can use distributed client or compute locally.

    Uses distributed resources parameter to control scheduling of GPUs.
    Pipeline produces jobs per DM/dt.
    Returns a dict with values as futures of certain jobs (data, collection).
    """

    logger.info('Building dask for observation {0}, scan {1}, segment {2}.'
                .format(st.metadata.datasetId, st.metadata.scan, segment))

    futures = {}

# new style read *TODO: note hack on resources*
    data = lazy_read_segment(st, segment, cfile, vys_timeout)
    data = cl.compute(data, resources={tuple(data.__dask_keys__()[0][0][0]):
                                       {'READER': 1}})  # get future
# old style read
#    data = cl.submit(source.read_segment, st, segment, cfile, vys_timeout,
#                     resources={'READER': 1})

    futures['data'] = data  # save future

    # TODO: put this on READER worker?
    data_prep = cl.submit(source.data_prep, st, segment, data,
                          resources={'CORES': st.prefs.nthread},
                          priority=1)

    saved = []
    if st.fftmode == "fftw":
        searchresources = {'CORES': st.prefs.nthread}
        imgranges = [[(min(st.get_search_ints(segment, dmind, dtind)),
                     max(st.get_search_ints(segment, dmind, dtind)))
                      for dtind in range(len(st.dtarr))]
                     for dmind in range(len(st.dmarr))]
        wisdom = cl.submit(search.set_wisdom, st.npixx, st.npixy,
                           resources={'CORES': 1})

        for dmind in range(len(st.dmarr)):
            delay = cl.submit(util.calc_delay, st.freq, st.freq.max(),
                              st.dmarr[dmind], st.inttime,
                              resources={'CORES': 1})
            for dtind in range(len(st.dtarr)):
                data_corr = cl.submit(search.dedisperseresample, data_prep,
                                      delay, st.dtarr[dtind],
                                      parallel=st.prefs.nthread > 1,
                                      resources={'CORES': st.prefs.nthread})

                im0, im1 = imgranges[dmind][dtind]
                integrationlist = [list(range(im0, im1)[i:i+st.chunksize])
                                   for i in range(0, im1-im0, st.chunksize)]
                for integrations in integrationlist:
                    saved.append(cl.submit(search.search_thresh_fftw, st,
                                           segment, data_corr, dmind, dtind,
                                           integrations=integrations,
                                           wisdom=wisdom,
                                           resources=searchresources))
        candcollection = cl.submit(mergelists, saved, resources={'CORES': 1},
                                   priority=2)

    elif st.fftmode == "cuda":
        for dmind in range(len(st.dmarr)):
            candcollection = cl.submit(search.dedisperse_image_cuda, st,
                                       segment, data_prep, dmind,
                                       resources={'GPU': 1,
                                                  'CORES': st.prefs.nthread},
                                       priority=2)

    futures['candcollection'] = candcollection
    # TODO: incorporate saving of candcollection (as in prep_and_search function)

    return futures


def pipeline_scan_delayed(st, segments=None, cl=None, host=None, cfile=None,
                          vys_timeout=vys_timeout_default):
    """ Submit pipeline processing of a single segment to scheduler.
    Uses delayed function and client.compute to schedule.

    Returns a list of dicts with futures of data, collection jobs.
    """

    if cl is None:
        if host is not None:
            cl = distributed.Client('{0}:{1}'.format(host, '8786'))

    if not isinstance(segments, list):
        segments = list(range(st.nsegment))

    futures = []
    for segment in segments:
        future = {}
        resources = {}

        logger.info('Building dask for observation {0}, scan {1}, segment {2}.'
                    .format(st.metadata.datasetId, st.metadata.scan, segment))

        # get data
        data = delayed(source.read_segment)(st, segment, cfile, vys_timeout)
        resources[tuple(data.__dask_keys__())] = {'READER': 1}

        # search data
        candcollection = delayed(search.prep_and_search)(st, segment, data)
        resources[tuple(candcollection.__dask_keys__())] = {'GPU': 1,
                                                            'CORES': st.prefs.nthread}

        # using submit with worker_client
#        data = cl.submit(read_segment, st, segment, cfile, vys_timeout,
#                         resources={'READER': 1})
#        candcollection = cl.submit(prep_and_search, st, segment, data,
#                                   resources={'GPU': 1})
#        future['data'] = data
#        future['candcollection'] = candcollection

        if cl is not None:
            future['data'] = cl.compute(data, resources=resources)
            future['candcollection'] = cl.compute(candcollection,
                                                  resources=resources)
            futures.append(future)
        else:
            futures.append(candcollection)

    return futures


### helper functions (not necessarily in use)

def read_segment(st, segment, cfile, vys_timeout):
    """ Wrapper for source.read_segment that secedes from worker
    thread pool
    """

    logger.info("Reading datasetId {0}, segment {1} locally."
                .format(st.metadata.scanId, segment))

    with distributed.worker_client() as cl_loc:
        fut = cl_loc.submit(source.read_segment, st, segment, cfile,
                            vys_timeout)
        data = fut.result()

    logger.info("Finished reading datasetId {0}, segment {1} locally."
                .format(st.metadata.scanId, segment))

    return data


def prep_and_search(st, segment, data):
    """ Wrapper for search.prep_and_search that secedes from worker
    thread pool
    """

    logger.info("Searching datasetId {0}, segment {1} locally."
                .format(st.metadata.scanId, segment))

    with distributed.worker_client() as cl_loc:
        fut = cl_loc.submit(search.prep_and_search, st, segment, data)
        cc = fut.result()

    logger.info("Finished searching datasetId {0}, segment {1} locally."
                .format(st.metadata.scanId, segment))

    return cc


def mergelists(futlists):
    """ Take list of lists and return single list
    ** TODO: could put logic here to find islands, peaks, etc?
    """

    return [fut for futlist in futlists for fut in futlist]


def lazy_read_segment(st, segment, cfile=None,
                      timeout=vys_timeout_default):
    """ rfpipe read_segment as a dask array.
    equivalent to making delayed version of function and then:
    arr = dask.array.from_delayed(dd, st.datashape, np.complex64).
    """

    shape = st.datashape
    chunks = ((shape[0],), (shape[1],), (shape[2],), (shape[3],))

    name = 'read_segment-' + tokenize([st, segment])
    dask = {(name, 0, 0, 0, 0): (source.read_segment, st, segment,
                                 cfile, timeout)}

    return array.Array(dask=dask, name=name, chunks=chunks, dtype=np.complex64)



from __future__ import print_function, division, absolute_import #, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open
from future.moves.urllib.parse import urlparse, urlunparse, urlencode
from future.moves.urllib.request import urlopen

import os.path
from lxml import etree, objectify
from astropy import time
from rfpipe.metadata import Metadata
import logging
logger = logging.getLogger(__name__)

_install_dir = os.path.abspath(os.path.dirname(__file__))
_xsd_dir = os.path.join(_install_dir, 'xsd')
_sdmbuilder_xsd = os.path.join(_xsd_dir, 'SdmBuilderMessage.xsd')
_sdmbuilder_parser = objectify.makeparser(
        schema=etree.XMLSchema(file=_sdmbuilder_xsd))

_host = 'mctest.evla.nrao.edu'
_path = 'sdm-builder/offline'


class SDMBuilder(object):
    """ Use mcaf to create new SDM from bdf
    """

    _E = objectify.ElementMaker(annotate=False)

    def __init__(self, datasetId=None, uid=None, dataSize=None,
                 numIntegrations=None, startTime=None, endTime=None,
                 host=_host, path=_path):
        self.datasetId = datasetId
        self.uid = uid
        self.dataSize = dataSize
        self.numIntegrations = numIntegrations
        self.startTime = startTime
        self.endTime = endTime
        self.host = host
        self.path = path

    @property
    def _root(self):
        return self._E.SdmBuilderMessage(
                self._E.datasetId(self.datasetId),
                self._E.bdf(
                    self._E.uid(self.uid),
                    self._E.dataSize(self.dataSize),
                    self._E.numIntegrations(self.numIntegrations),
                    self._E.startTime(repr(self.startTime)),
                    self._E.endTime(repr(self.endTime)),
                    ),
                # SdmBuilderMessage attributes:
                {'timestamp': '%.12f' % time.Time.now().mjd,
                    'sender': 'realfast'}
                )

    @property
    def xml(self):
        return etree.tostring(self._root, xml_declaration=True,
                              pretty_print=False, standalone=True)

    @property
    def _url(self):
        query = urlencode({'xml': self.xml})
        url = urlunparse(('https', self.host, self.path, '', query, ''))
        return url

    def send(self):
        response_xml = urlopen(self._url).read()
        if b'error' in response_xml:
            self.response = None
        else:
            self.response = objectify.fromstring(response_xml,
                                                 parser=_sdmbuilder_parser)

    @property
    def location(self):
        try:
            return str(self.response.result.location)
        except AttributeError:
            logger.warn("No SDM generated.")
            return None


def makesdm(startTime, endTime, metadata, data):
    """ Generates call to sdm builder server for a single candidate.
    Generates a unique id for the bdf from the startTime.
    Uses metadata and data to create call signature to server with:
    (datasetId, dataSize_bytes, nint, startTime_mjd, endTime_mjd)
    Returns location of newly created SDM.
    Data refers to cut out visibilities from startTime to endTime with
    shape of (nint, nbl, nspw, numBin, nchan, npol).
    """

    assert type(metadata) == Metadata, ("metadata must be "
                                        "of type rfpipe.metadata.Metadata")

    assert data.ndim == 6, ("data must have 6 dimensions: "
                            "nint, nbl, nspw, numBin, nchan, npol.")

    nint, nbl, nspw, numBin, nchan, npol = data.shape
    dataSize = data.nbytes
    uid = ('uid:///evla/realfastbdf/{0}'
           .format(int(time.Time(startTime, format='mjd').unix*1e3)))
    logger.info("Building SDM for datasetId {0} and bdf {1}"
                .format(metadata.datasetId, uid))
    sdmb = SDMBuilder(metadata.datasetId, uid, dataSize, nint, startTime,
                      endTime)
    sdmb.send()

    return sdmb.location


def makebdf(startTime, endTime, metadata, data, bdfdir='.'):
    """ Create bdf for candidate that contains data array.
    Data is numpy array of complex64 type spanning start/endTime.
    Should have shape (nint, nbl, nspw, nbin, nchan, npol).
    metadata is a rfpipe.metadata.Metadata object.
    Assumes one bdf per sdm and one sdm per candidate.
    Only supports 8bit samplers and IFid of AC/BD.
    """

    from sdmpy import bdf

    assert type(metadata) == Metadata, ("metadata must be "
                                        "of type rfpipe.metadata.Metadata")

    assert data.ndim == 6, ("data must have 6 dimensions: "
                            "nint, nbl, nspw, numBin, nchan, npol.")

    nint, nbl, nspw, numBin, nchan, npol = data.shape

    IFidspwnum = [spw.split('-') for (spw, freq) in metadata.spworder]
    spws = [bdf.BDFSpectralWindow(None, numBin=numBin, numSpectralPoint=nchan,
                                  sw=int(swnum)+1,   # casting from 0->1 based
                                  swbb='{0}_8BIT'.format(IFid),
                                  npol=npol) for (IFid, swnum) in IFidspwnum]
    # TODO: confirm that sw is 1 based for a proper SDM
    # TODO: confirm that metadata spworder is 0 based

    assert nspw == len(spws), ('Expected one spw in metadata.spworder per spw '
                               'in data array.')
    assert os.path.isdir(bdfdir), 'bdfdir does not exist'

    uid = ('uid:///evla/realfastbdf/{0}'
           .format(int(time.Time(startTime, format='mjd').unix*1e3)))
    w = bdf.BDFWriter(bdfdir, start_mjd=startTime, uid=uid,
                      num_antenna=metadata.nants_orig, spws=spws, scan_idx=1,
                      corr_mode='c')

    dat = {}
    w.write_header()
    for i in range(nint):
        dat['crossData'] = data[i]
        ts = startTime+metadata.inttime/2/86400.
        w.write_integration(mjd=ts, interval=metadata.inttime, data=dat)
    w.close()



from __future__ import print_function, division, absolute_import #, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

import os.path
from elasticsearch import Elasticsearch, RequestError, TransportError, helpers
import pickle
import logging
logging.getLogger('elasticsearch').setLevel(30)
logger = logging.getLogger(__name__)
logger.setLevel(10)

# eventually should be updated to search.realfast.io/api with auth
es = Elasticsearch(['go-nrao-nm.aoc.nrao.edu:9200'])


def indexscan_config(config, preferences=None, datasource='vys',
                     indexprefix='new'):
    """ Takes scan config and creates dict to push
    to elasticsearch scnas index.
    Optionally pushes rfpipe preferences object as separate doc
    and connects them via the hexdigest name.
    Note index names must end in s and types are derived as singular form.
    datasource is assumed to be 'vys', but 'sim' is an option.
    indexprefix allows specification of set of indices ('test', 'aws').
    Use indexprefix='new' for production.
    """

    scandict = {}

    # define dict for scan properties to index
    scandict['datasetId'] = config.datasetId
    scandict['scanId'] = config.scanId
    scandict['projid'] = config.projid
    scandict['scanNo'] = int(config.scanNo)
    scandict['subscanNo'] = int(config.subscanNo)
    scandict['source'] = str(config.source)
    scandict['ra_deg'] = float(config.ra_deg)
    scandict['dec_deg'] = float(config.dec_deg)
    scandict['startTime'] = float(config.startTime)
    scandict['stopTime'] = float(config.stopTime)
    scandict['datasource'] = datasource
    scandict['scan_intent'] = config.scan_intent

    # if preferences provided, it will connect them by a unique name
    if preferences:
        scandict['prefsname'] = preferences.name
        indexprefs(preferences, indexprefix=indexprefix)

    # push scan info with unique id of scanId
    index = indexprefix+'scans'
    res = pushdata(scandict, index=index, Id=config.scanId,
                   command='index')
    if res == 1:
        logger.info('Indexed scan config {0} to {1}'.format(config.scanId,
                                                            index))
    else:
        logger.warn('Scan config not indexed for {0}'.format(config.scanId))


def indexscan_sdm(sdmfile, sdmscan, sdmsubscan, preferences=None,
                  datasource='sdm', indexprefix='new'):
    """ Takes sdmfile and sdmscan and pushes to elasticsearch scans index.
    indexprefix allows specification of set of indices ('test', 'aws').
    Use indexprefix='new' for production.
    """

# must include:
#    scanproperties = ['datasetId', 'scanNo', 'subscanNo', 'projid', 'ra_deg',
#                      'dec_deg', 'scan_intent', 'source', 'startTime',
#                      'stopTime']

    import sdmpy
    from numpy import degrees

    datasetId = os.path.basename(sdmfile.rstrip('/'))
    scanId = '{0}.{1}.{2}'.format(datasetId, str(sdmscan), str(sdmsubscan))

    sdm = sdmpy.SDM(sdmfile, use_xsd=False)
    scan = sdm.scan(sdmscan)

    scandict = {}

    # define dict for scan properties to index
    scandict['datasetId'] = datasetId
    scandict['scanId'] = scanId
    scandict['projid'] = 'Unknown'
    scandict['scanNo'] = int(sdmscan)
    scandict['subscanNo'] = int(sdmsubscan)
    scandict['source'] = str(scan.source)
    ra, dec = degrees(scan.coordinates)
    scandict['ra_deg'] = float(ra)
    scandict['dec_deg'] = float(dec)
    scandict['startTime'] = float(scan.startMJD)
    scandict['stopTime'] = float(scan.endMJD)
    scandict['datasource'] = datasource
    scandict['scan_intent'] = ','.join(scan.intents)

    # if preferences provided, it will connect them by a unique name
    if preferences:
        scandict['prefsname'] = preferences.name
        indexprefs(preferences, indexprefix=indexprefix)

    # push scan info with unique id of scanId
    index = indexprefix+'scans'
    res = pushdata(scandict, index=index, Id=scanId,
                   command='index')
    if res == 1:
        logger.info('Indexed scan config {0} to {1}'.format(scanId, index))
    else:
        logger.warn('Scan config not indexed for {0}'.format(scanId))


def indexscan_meta(metadata, preferences=None, indexprefix='new'):
    """ Takes metadata object and pushes to elasticsearch scans index.
    indexprefix allows specification of set of indices ('test', 'aws').
    Use indexprefix='new' for production.
    """

# must include:
#    scanproperties = ['datasetId', 'scanNo', 'subscanNo', 'projid', 'ra_deg',
#                      'dec_deg', 'scan_intent', 'source', 'startTime',
#                      'stopTime']

    from numpy import degrees

    scandict = {}

    # define dict for scan properties to index
    scandict['datasetId'] = metadata.datasetId
    scandict['scanId'] = metadata.scanId
    scandict['projid'] = 'Unknown'
    scandict['scanNo'] = int(metadata.scan)
    scandict['subscanNo'] = int(metadata.subscan)
    scandict['source'] = metadata.source
    ra, dec = degrees(metadata.radec)
    scandict['ra_deg'] = float(ra)
    scandict['dec_deg'] = float(dec)
    scandict['startTime'] = float(metadata.starttime_mjd)
    scandict['stopTime'] = float(metadata.endtime_mjd)
    scandict['datasource'] = metadata.datasource
    scandict['scan_intent'] = metadata.intent  # assumes ,-delimited string

    # if preferences provided, it will connect them by a unique name
    if preferences:
        scandict['prefsname'] = preferences.name
        indexprefs(preferences, indexprefix=indexprefix)

    # push scan info with unique id of scanId
    index = indexprefix+'scans'
    res = pushdata(scandict, index=index, Id=metadata.scanId,
                   command='index')
    if res == 1:
        logger.info('Indexed scan config {0} to {1}'
                    .format(metadata.scanId, index))
    else:
        logger.warn('Scan config not indexed for {0}'.format(metadata.scanId))


def indexprefs(preferences, indexprefix='new'):
    """ Index preferences with id equal to hash of contents.
    indexprefix allows specification of set of indices ('test', 'aws').
    Use indexprefix='new' for production.
    """

    index = indexprefix+'preferences'
    res = pushdata(preferences.ordered, index=index,
                   Id=preferences.name, command='index')
    if res == 1:
        logger.info('Indexed preference {0} to {1}'
                    .format(preferences.name, index))
    else:
        logger.warn('Preferences not indexed for {0}'.format(preferences.name))


def indexcands(candcollection, scanId, tags=None, url_prefix=None,
               indexprefix='new'):
    """ Takes candidate collection and pushes to index
    Connects to preferences via hashed name
    scanId is added to associate cand to a give scan.
    Assumes scanId is defined as:
    datasetId dot scanNo dot subscanNo.
    tags is a comma-delimited string used to fill tag field in index.
    indexprefix allows specification of set of indices ('test', 'aws').
    Use indexprefix='new' for production.
    """

    if tags is None:
        tags = ''

    index = indexprefix+'cands'

    # create new tag string with standard format to fill in blanks
    allowed_tags = ["rfi", "bad", "noise", "interesting", "astrophysical",
                    "mock"]
    tagstr = ','.join([tag for tag in tags.split(',') if tag in allowed_tags])

    candarr = candcollection.array
    prefs = candcollection.prefs
    candmjd = candcollection.candmjd
    canddm = candcollection.canddm
    canddt = candcollection.canddt

    res = 0
    for i in range(len(candarr)):
        # get features. use .item() to cast to default types
        canddict = dict(list(zip(candarr.dtype.names, candarr[i].item())))

        # fill optional fields
        canddict['scanId'] = scanId
        datasetId, scan, subscan = scanId.rsplit('.', 2)
        canddict['datasetId'] = datasetId
        canddict['scan'] = int(scan)
        canddict['subscan'] = int(subscan)
        canddict['tags'] = tagstr
        canddict['tagcount'] = 0
        canddict['candmjd'] = float(candmjd[i])
        canddict['canddm'] = float(canddm[i])
        canddict['canddt'] = float(canddt[i])
        canddict['png_url'] = ''
        if prefs.name:
            canddict['prefsname'] = prefs.name

        # create id
        uniqueid = candid(canddict)
        candidate_png = 'cands_{0}.png'.format(uniqueid)
        if os.path.exists(candidate_png):
                logger.debug("Found png {0} and setting cands index field"
                             .format(candidate_png))
                canddict['png_url'] = os.path.join(url_prefix, candidate_png)

        res += pushdata(canddict, index=index,
                        Id=uniqueid, command='index')

    if res >= 1:
        logger.debug('Indexed {0} cands for {1} to {2}'.format(res, scanId,
                                                               index))
    else:
        logger.debug('No cands indexed for {0}'.format(scanId))

    return res


def indexmocks(st, indexprefix='new'):
    """ Reads simulated_transient from state and pushes to index
    Mock index must include scanId to reference data that was received.
    indexprefix allows specification of set of indices ('test', 'aws').
    Use indexprefix='new' for production.
    """

    if st.prefs.simulated_transient is None:
        return 0

    mocks = st.prefs.simulated_transient
    scanId = st.metadata.scanId
    index = indexprefix+'mocks'

    res = 0
    for mock in mocks:
        mockdict = {}
        mockdict['scanId'] = scanId
        (seg, i0, dm, dt, amp, l, m) = mock
        mockdict['segment'] = int(seg)
        mockdict['integration'] = int(i0)
        mockdict['dm'] = float(dm)
        mockdict['dt'] = float(dt)
        mockdict['amp'] = float(amp)
        mockdict['l'] = float(l)
        mockdict['m'] = float(m)

        res += pushdata(mockdict, Id=scanId, index=index,
                        command='index')

    if res >= 1:
        logger.debug('Indexed {0} mocks for {1} to {2}'.format(res, scanId,
                                                               index))
    else:
        logger.debug('No mocks indexed for {0}'.format(scanId))

    return res


def indexnoises(noisefile, scanId, indexprefix='new'):
    """ Reads noises from noisefile and pushes to index
    scanId is added to associate cand to a give scan.
    indexprefix allows specification of set of indices ('test', 'aws').
    Use indexprefix='' for production.
    """

    index = indexprefix+'noises'
    doc_type = index.rstrip('s')

    noises = []
    with open(noisefile, 'rb') as pkl:
        noises += pickle.load(pkl)

    count = 0
    for noise in noises:
        segment, integration, noiseperbl, zerofrac, imstd = noise
        Id = '{0}.{1}.{2}'.format(scanId, segment, integration)
        if not es.exists(index=index, doc_type=doc_type, id=Id):
            noisedict = {}
            noisedict['scanId'] = str(scanId)
            noisedict['segment'] = int(segment)
            noisedict['integration'] = int(integration)
            noisedict['noiseperbl'] = float(noiseperbl)
            noisedict['zerofrac'] = float(zerofrac)
            noisedict['imstd'] = float(imstd)

            count += pushdata(noisedict, Id=Id, index=index,
                              command='index')

    if count:
        logger.debug('Indexed {0} noises for {1} to {2}'
                     .format(count, scanId, index))
    else:
        logger.debug('No noises indexed for {0}'.format(scanId))

    return count


def pushdata(datadict, index, Id=None, command='index', force=False):
    """ Pushes dict to index, which can be:
    candidates, scans, preferences, or noises
    Assuming one elasticsearch doc_type per index (less the s)
    Id for scan should be scanId, while for preferences should be hexdigest
    Command can be 'index' or 'delete'.
    To update, index with existing key and force=True.
    """

    assert isinstance(datadict, dict)

    # only one doc_type per index and its name is derived from index
    doc_type = index.rstrip('s')

    logger.debug('Pushing to index {0} with Id {1}'.format(index, Id))
    res = 0

    if command == 'index':
        if force:
            res = es.index(index=index, doc_type=doc_type, id=Id,
                           body=datadict)
        else:
            if not es.exists(index=index, doc_type=doc_type, id=Id):
                try:
                    res = es.index(index=index, doc_type=doc_type,
                                   id=Id, body=datadict)
                except RequestError:
                    logger.warn("Id {0} and data {1} not indexed due to request error."
                                .format(Id, datadict))
            else:
                logger.warn('Id={0} already exists in index {1}'
                            .format(Id, index))

    elif command == 'delete':
        if es.exists(index=index, doc_type=doc_type, id=Id):
            res = es.delete(index=index, doc_type=doc_type, id=Id)
        else:
            logger.warn('Id={0} not in index'.format(Id))

    if res:
        return res['_shards']['successful']
    else:
        return res


def candid(data):
    """ Returns id string for given data dict
    Assumes scanId is defined as:
    datasetId dot scanNum dot subscanNum
    """

    return ('{0}_seg{1}-i{2}-dm{3}-dt{4}'
            .format(data['scanId'], data['segment'], data['integration'],
                    data['dmind'], data['dtind']))


def get_ids(index, **kwargs):
    """ Gets Ids from an index
    doc_type derived from index name (one per index)
    Can optionally pass key-value pairs of field-string to search.
    Must match exactly (e.g., "scanId"="test.1.1")
    """

    # only one doc_type per index and its name is derived from index
    doc_type = index.rstrip('s')
    if 'field' in kwargs:
        field = kwargs.pop('field')
    else:
        field = 'false'

    if len(kwargs):
        query = {"query": {"match": kwargs}, "_source": field}
    else:
        query = {"query": {"match_all": {}}, "_source": field}

    res = helpers.scan(es, index=index, doc_type=doc_type, query=query)

    if field == 'false':
        return [hit['_id'] for hit in res]
    else:
        return [(hit['_id'], hit['_source'][field]) for hit in res]


def update_field(index, field, value, **kwargs):
    """ Replace an index's field with a value.
    Optionally query the index with kwargs.
    Use with caution.
    """

    doc_type = index.rstrip('s')
    if len(kwargs):
        searchquery = {"match": kwargs}
    else:
        searchquery = {"match_all": {}}

    query = {"query": searchquery,
             "script": {"inline": "ctx._source.{0}='{1}'".format(field, value),
                        "lang": "painless"}}

    resp = es.update_by_query(body=query, doc_type=doc_type, index=index)

    response_info = {"total": resp["total"], "updated": resp["updated"],
                     "type": "success"}
    if resp["failures"] != []:
        response_info["type"] = "failure"

    return response_info


def remove_ids(index, **kwargs):
    """ Gets Ids from an index
    doc_type derived from index name (one per index)
    Can optionally pass key-value pairs of field-string to search.
    Must match exactly (e.g., "scanId"="test.1.1")
    """

    if not len(kwargs):
        logger.warn("No query kwargs provided. Will clear all Ids in {0}"
                    .format(index))
        confirm = input("Press any key to confirm removal.")
        if confirm:
            logger.info("Removing...")

    Ids = get_ids(index, **kwargs)
    res = 0
    for Id in Ids:
        res += pushdata({}, index, Id, command='delete')

    logger.info("Removed {0} docs from index {1}".format(res, index))


def create_indices(indexprefix):
    """ Create standard set of indices,
    cands, scans, preferences, mocks, noises
    """

    body = {"settings": {
                "analysis": {
                    "analyzer": {
                        "default": {"tokenizer": "whitespace"}
                        }
                    }
                },
            }

    body_preferences = body.copy()
    body_preferences['mappings'] = {indexprefix+"preference": {
                                     "properties": {
                                       "flaglist": {"type":  "text"}
                                       }
                                     }
                                    }

    indices = ['scans', 'cands', 'preferences', 'mocks', 'noises']
    for index in indices:
        fullindex = indexprefix+index
        if es.indices.exists(index=fullindex):
            confirm = input("Index {0} exists. Delete?".format(fullindex))
            if confirm:
                es.indices.delete(index=fullindex)
        if index != 'preferences':
            es.indices.create(index=fullindex, body=body)
        else:
            es.indices.create(index=fullindex, body=body_preferences)


def reset_indices(indexprefix, deleteindices=False):
    """ Remove entries from set of indices with a given indexprefix.
    indexprefix allows specification of set of indices ('test', 'aws').
    Use indexprefix='new' for production.
    deleteindices will delete indices, too.

    *BE SURE YOU KNOW WHAT YOU ARE DOING*
    """

    logger.warn("Erasing all docs from indices with prefix {0}"
                .format(indexprefix))

    for index in [indexprefix+'noises', indexprefix+'mocks',
                  indexprefix+'cands', indexprefix+'scans',
                  indexprefix+'preferences']:
        confirm = input("Confirm removal of {0} entries"
                        .format(index))
        res = 0
        if confirm:
            try:
                Ids = get_ids(index)
                for Id in Ids:
                    res += pushdata({}, index, Id, command='delete')
                logger.info("Removed {0} docs from index {1}".format(res, index))
                if deleteindices:
                    es.indices.delete(index)
                    logger.info("Removed {0} index".format(index))
            except TransportError:
                logger.info("Index {0} does not exist".format(index))


def move_doc(index1, index2, Id, deleteorig=True):
    """ Take doc in index1 with Id and move to index2
    """

    assert 'final' not in index1

    doc_type1 = index1.rstrip('s')
    doc_type2 = index2.rstrip('s')

    doc = es.get(index=index1, doc_type=doc_type1, id=Id)
    res = es.index(index=index2, doc_type=doc_type2, id=Id,
                   body=doc['_source'])

    if res['_shards']['successful']:
        if deleteorig:
            es.delete(index=index1, doc_type=doc_type1, id=Id)
    else:
        logger.warn("Move of {0} from index {1} to {2} failed".format(Id,
                                                                      index1,
                                                                      index2))

    return res['_shards']['successful']


def find_docids(indexprefix, candId=None, scanId=None):
    """ Use id from one index to find ids of others (like a relational db).
    Returns a dict with keys of the index name and values of the related ids.
    A full index set has:
        - cands indexed by candId (has scanId field)
        - scans indexed by scanId
        - preferences indexed by preferences name (in scans index)
        - mocks indexed by scanId (has scanId field)
        - noises indexed by noiseId (has scanId field)
    """

    docids = {}

    if candId is not None and scanId is None:
        index = indexprefix + 'cands'
        doc_type = index.rstrip('s')

        doc = es.get(index=index, doc_type=doc_type, id=candId)

    if scanId is not None and candId is None:
        # use scanId to get ids with one-to-many mapping
        for ind in ['cands', 'mocks', 'noises']:
            index = indexprefix + ind
            ids = get_ids(index, _id=scanId)
            docids[index] = ids

        # get prefsname from scans index
        index = indexprefix + 'scans'
        docids[index] = [scanId]
        _, prefsname = get_ids(index, _id=scanId, field='prefsname')[0]
        index = indexprefix + 'preferences'
        docids[index] = [prefsname]

    return docids


def move_docs(indexprefix1, indexprefix2, candids):
    """ Given candids, move *all* docs from indexprefix1 to indexprefix2.
    This will find all 
    """

    for Id in candids:
        docids = find_docids(indexprefix=indexprefix1, candId=Id)
        for index, Id0 in docids:
            index2 = indexprefix2 + index.lstrip(indexprefix1)
            move_doc(index, index2, Id0)

    # TODO: don't move preferences, since they may have other scanids associated with them. Just copy.


def find_consensus(indexprefix='new', nop=3):
    """ Pull candidates with at least nop user tag fields to find consensus.
    """

    # should copy existing tags and user tags fields
    # also should update tags field with action on underlying data

    assert indexprefix != 'final'
    index = indexprefix+'cands'
    doc_type = index.rstrip('s')

    ids = get_ids(index, tagcount=nop)

    for Id in ids:
        doc = es.get(index=index, doc_type=doc_type, id=Id)
        tags = dict(((k, v) for (k, v) in doc['_source'].items() if '_tags' in k))
        logger.info("Id {0} has {1} tags: {2}".format(Id, len(tags), tags))
#       build dict of ids with agreeing tags and disagreeing tags



from __future__ import print_function, division, absolute_import#, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

from rfpipe import state
import os.path
from math import log
import logging
logger = logging.getLogger(__name__)


def reader_memory_available(cl):
    """ Calc memory in use by READERs
    """

    return [vals['memory_limit']-vals['memory']
            for vals in itervalues(cl.scheduler_info()['workers'])
            if 'READER' in vals['resources']]


def reader_memory_used(cl):
    """ Calc memory in use by READERs
    """

    return [vals['memory']
            for vals in itervalues(cl.scheduler_info()['workers'])
            if 'READER' in vals['resources']]


def spilled_memory(daskdir='/lustre/evla/test/realfast/dask-worker-space'):
    """ How much memory has been spilled by dask/distributed?
    """

    spilled = 0
    for dirpath, dirnames, filenames in os.walk(daskdir):
        for filename in filenames:
            try:
                spilled += os.path.getsize(os.path.join(dirpath, filename))/1024.**3
            except OSError:
                try:
                    spilled += os.path.getsize(os.path.join(dirpath, filename))/1024.**3
                except OSError:
                    logger.warn("Could not get size of spilled file. Skipping.")

    return spilled


def reader_memory_ok(cl, memory_required):
    """ Does any READER worker have enough memory?
    memory_required is the size of the read in bytes
    """

    for worker_memory in reader_memory_available(cl):
        if worker_memory > memory_required:
            return True

    logger.info("No worker found with required memory of {0} GB"
                .format(memory_required/1e9))

    return False


def readertotal_memory_ok(cl, memory_limit):
    """ Is total READER memory usage too high?
    memory_limit is total memory used in bytes
    """

    if memory_limit is not None:
        total = sum(reader_memory_used(cl))

        if total > memory_limit:
            logger.info("Total of {0} GB in use. Exceeds limit of {1} GB."
                        .format(total/1e9, memory_limit/1e9))

        return total < memory_limit
    else:
        return True


def spilled_memory_ok(limit=1.0, daskdir='/lustre/evla/test/realfast/dask-worker-space'):
    """ Calculate total memory spilled (in GB) by dask distributed.
    """

    spilled = spilled_memory(daskdir)

    if spilled < limit:
        return True
    else:
        logger.info("Spilled memory {0:.1f} GB exceeds limit of {1:.1f}"
                    .format(spilled, limit))
        return False


def valid_telcalfile(st):
    """ Test whether telcalfile exists at the moment.
    Note: telcalfile may appear later.
    """

    if os.path.exists(st.gainfile) and os.path.isfile(st.gainfile):
        return True
    else:
        return False


def state_validates(config=None, inmeta=None, sdmfile=None, sdmscan=None,
                   bdfdir=None, preffile=None, inprefs={}):
    """ Try to compile state
    """

    try:
        st = state.State(inmeta=inmeta, config=config, preffile=preffile,
                         inprefs=inprefs, sdmfile=sdmfile, sdmscan=sdmscan,
                         bdfdir=bdfdir, showsummary=False, validate=True)
        return True
    except AssertionError:
        return False


def total_images_searched(st):
    """ Number of images formed (trials) in all segments, dms, dts.
    """

    si = 0
    for segment in range(st.nsegment):
        for dmind in range(len(st.dmarr)):
            for dtind in range(len(st.dtarr)):
                si += len(st.get_search_ints(segment, dmind, dtind))
    return si


def total_compute_time(st):
    """ Uses a simple model for total GPU compute time (in sec) based on profiling.
    Models the GPU time per trial (incl data in, amortized over many dm/dt).
    No distributed data movement time included.
    2.3e-4 s (512x512)
    6.1e-4 s (1024x1024)
    1.2e-3 s (2048x2048)
    3.8e-3 s (4096x4096)
    """

    time_ref = 2.3e-4
    npix_ref = 512

    si = total_images_searched(st)
    npix = (st.npixx+st.npixy)/2

    return si * time_ref * npix*log(npix)/(npix_ref*log(npix_ref))


def total_memory_read(st):
    """ Memory read (in GB) including overlapping read at segment boundaries.
    """

    return st.nsegment*st.vismem



from __future__ import print_function, division, absolute_import#, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

import pickle
import os.path
import glob
import shutil
import math
import random
import distributed
from astropy import time
from time import sleep
import dask.utils
from evla_mcast.controller import Controller
from rfpipe import state, preferences, candidates
from realfast import pipeline, elastic, sdm_builder, heuristics

import logging
import matplotlib
import yaml
matplotlib.use('Agg')
ch = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s %(levelname)8s %(name)s | %(message)s')
ch.setFormatter(formatter)
logger = logging.getLogger('realfast_controller')

_vys_cfile_prod = '/home/cbe-master/realfast/lustre_workdir/vys.conf'  # production file
_vys_cfile_test = '/home/cbe-master/realfast/soft/vysmaw_apps/vys.conf'  # test file
_preffile = '/lustre/evla/test/realfast/realfast.yml'
_distributed_host = '192.168.201.101'  # for ib0 on cbe-node-01
_candplot_dir = '/users/claw/public_html/realfast/plots'
_candplot_url_prefix = 'http://www.aoc.nrao.edu/~claw/realfast/plots'


class realfast_controller(Controller):

    def __init__(self, preffile=None, inprefs={}, host=None, **kwargs):
        """ Creates controller object that can act on a scan configuration.
        Inherits a "run" method that starts asynchronous operation that calls
        handle_config. handle_sdm and handle_meta (with datasource "vys" or
        "sim") are also supported.
        host allows specification of 'localhost' for distributed client.

        kwargs can include:
        - tags, a comma-delimited string for cands to index
        - nameincludes, a string required to be in datasetId,
        - vys_timeout, factor over real-time for vys reading to wait,
        - vys_sec_per_spec, time in sec to allow for vys reading (overloaded by vys_timeout)
        - mockprob, chance (range 0-1) that a mock is added to scan,
        - saveproducts, boolean defining generation of mini-sdm,
        - indexresults, boolean defining push (meta)data to search index,
        - archiveproducts, boolean defining archiving mini-sdm,
        - throttle, boolean defining whether to slow pipeline submission,
        - read_overhead, throttle param requires multiple of vismem in a READERs memory,
        - read_totfrac, throttle param requires fraction of total READER memory be available,
        - spill_limit, throttle param limiting maximum size (in GB) of data spill directory,
        - searchintents, a list of intent names to search,
        - indexprefix, a string defining set of indices to save results.
        """

        super(realfast_controller, self).__init__()

        self.inprefs = inprefs  # rfpipe preferences
        if host is None:
            self.client = distributed.Client('{0}:{1}'
                                             .format(_distributed_host,
                                                     '8786'))
        elif host == 'localhost':
            self.client = distributed.Client(n_workers=1,
                                             threads_per_worker=4,
                                             resources={"READER": 1,
                                                        "CORES": 4})

        self.lock = dask.utils.SerializableLock()
        self.states = {}
        self.futures = {}
        self.futures_removed = {}

        # define attributes from yaml file
        self.preffile = preffile if preffile is not None else _preffile
        prefs = {}
        if os.path.exists(self.preffile):
            with open(self.preffile, 'r') as fp:
                prefs = yaml.load(fp)['realfast']
                logger.info("Parsed realfast preferences from {0}"
                            .format(self.preffile))
        else:
            logger.warn("realfast preffile {0} given, but not found"
                        .format(self.preffile))

        # get arguments from preffile, optional overload from kwargs
        self.daskdir = '/lustre/evla/test/realfast/dask-worker-space'
        for attr in ['tags', 'nameincludes', 'mockprob', 'vys_timeout',
                     'vys_sec_per_spec', 'indexresults', 'saveproducts',
                     'archiveproducts', 'searchintents', 'throttle',
                     'read_overhead', 'read_totfrac', 'spill_limit',
                     'indexprefix']:
            setattr(self, attr, None)
            if attr in prefs:
                setattr(self, attr, prefs[attr])
            if attr in kwargs:
                setattr(self, attr, kwargs[attr])

        if self.indexprefix is None:
            self.indexprefix = 'new'
        assert self.indexprefix in ['new', 'test', 'aws'], "indexprefix must be None, 'new', 'test' or 'aws'."

    def __repr__(self):
        return ('realfast controller with {0} jobs'
                .format(len(self.futures)))

    @property
    def statuses(self):
        return ['{0}, {1}: {2}'.format(scanId, ftype,
                                       futures[ftype].status)
                for (scanId, futurelist) in iteritems(self.futures)
                for futures in futurelist
                for ftype in futures]

    @property
    def errors(self):
        return ['{0}, {1}: {2}'.format(scanId, ftype,
                                       futures[ftype].exception())
                for (scanId, futurelist) in iteritems(self.futures)
                for futures in futurelist
                for ftype in futures
                if futures[ftype].status == 'error']

    @property
    def reader_memory_available(self):
        return heuristics.reader_memory_available(self.client)

    @property
    def reader_memory_used(self):
        return heuristics.reader_memory_used(self.client)

    @property
    def spilled_memory(self):
        return heuristics.spilled_memory(self.daskdir)

    def handle_config(self, config, cfile=_vys_cfile_prod, segments=None):
        """ Triggered when obs comes in.
        Downstream logic starts here.
        Default vys config file uses production parameters.
        segments arg can be used to submit a subset of all segments.

        """

        summarize(config)

        if search_config(config, nameincludes=self.nameincludes,
                         searchintents=self.searchintents):
            logger.info('Config looks good. Generating rfpipe state...')

            self.set_state(config.scanId, config=config, inmeta={'datasource':
                                                                 'vys'})
            self.inject_transient(config.scanId)  # randomly inject mock transient

            if self.indexresults:
                elastic.indexscan_config(config,
                                         preferences=self.states[config.scanId].prefs,
                                         datasource='vys',
                                         indexprefix=self.indexprefix)
            else:
                logger.info("Not indexing config or prefs.")

            self.start_pipeline(config.scanId, cfile=cfile, segments=segments)

        else:
            logger.info("Config not suitable for realfast. Skipping.")

        self.cleanup()

    def handle_sdm(self, sdmfile, sdmscan, bdfdir=None, segments=None):
        """ Parallel to handle_config, but allows sdm to be passed in.
        segments arg can be used to submit a subset of all segments.
        """

        # TODO: subscan assumed = 1
        sdmsubscan = 1
        scanId = '{0}.{1}.{2}'.format(os.path.basename(sdmfile.rstrip('/')),
                                      str(sdmscan), str(sdmsubscan))


        self.set_state(scanId, sdmfile=sdmfile, sdmscan=sdmscan, bdfdir=bdfdir,
                       inmeta={'datasource': 'sdm'})
        self.inject_transient(scanId)  # randomly inject mock transient

        if self.indexresults:
            elastic.indexscan_sdm(sdmfile, sdmscan, sdmsubscan,
                                  preferences=self.states[scanId].prefs,
                                  datasource='sdm',
                                  indexprefix=self.indexprefix)
        else:
            logger.info("Not indexing sdm scan or prefs.")

        self.start_pipeline(scanId, segments=segments)

        self.cleanup()

    def handle_meta(self, inmeta, cfile=_vys_cfile_test, segments=None):
        """ Parallel to handle_config, but allows metadata dict to be passed in.
        Gets called explicitly.
        Default vys config file uses test parameters.
        inmeta datasource key can be 'vys' or 'sim' and is passed to rfpipe.
        segments arg can be used to submit a subset of all segments.
        """

        scanId = '{0}.{1}.{2}'.format(inmeta['datasetId'], str(inmeta['scan']),
                                      str(inmeta['subscan']))

        self.set_state(scanId, inmeta=inmeta)
        self.inject_transient(scanId)  # randomly inject mock transient

        if self.indexresults:
            elastic.indexscan_meta(self.states[scanId].metadata,
                                   preferences=self.states[scanId].prefs,
                                   indexprefix=self.indexprefix)
        else:
            logger.info("Not indexing sdm scan or prefs.")

        self.start_pipeline(scanId, cfile=cfile, segments=segments)

        self.cleanup()

    def set_state(self, scanId, config=None, inmeta=None, sdmfile=None,
                  sdmscan=None, bdfdir=None, prefsname=None):
        """ Given metadata source, define state for a scanId.
        """

        # TODO: define prefsname according to config and/or heuristics
        prefs = preferences.Preferences(**preferences.parsepreffile(self.preffile,
                                                                    name=prefsname,
                                                                    inprefs=self.inprefs))

        st = state.State(inmeta=inmeta, config=config, inprefs=prefs,
                         lock=self.lock, sdmfile=sdmfile, sdmscan=sdmscan,
                         bdfdir=bdfdir)

        logger.info('State set for scanId {0}. Requires {1:.1f} GB read and'
                    ' {2:.1f} GPU-sec to search.'
                    .format(st.metadata.scanId,
                            heuristics.total_memory_read(st),
                            heuristics.total_compute_time(st)))

        self.states[scanId] = st

    def start_pipeline(self, scanId, cfile=None, segments=None):
        """ Start pipeline conditional on cluster state.
        Sets futures and state after submission keyed by scanId.
        segments arg can be used to select or slow segment submission.
        """

        st = self.states[scanId]

        vys_timeout = self.vys_timeout
        if st.metadata.datasource == 'vys':
            if self.vys_timeout is not None:
                logger.info("vys_timeout factor set to fixed value of {0:.1f}x"
                            .format(vys_timeout))
            else:
                assert self.vys_sec_per_spec is not None, "Must define vys_sec_per_spec to estimate vys_timeout"
                nspec = st.readints*st.nbl*st.nspw*st.npol
                vys_timeout = 1 + self.vys_sec_per_spec*nspec
                logger.info("vys_timeout factor scaled by nspec to {0:.1f}x"
                            .format(vys_timeout))

        if not heuristics.valid_telcalfile(st):
            logger.warn("telcalfile {0} for scanId {1} is not available at "
                        "pipeline submission"
                        .format(st.gainfile, scanId))

        # submit, with optional throttling
        futures = []
        if self.throttle and self.read_overhead and self.read_totfrac and self.spill_limit:
            logger.info('Starting pipeline throttled by read_overhead {0}, '
                        'read_totfrac {1}, and spill_limit {2}'
                        .format(self.read_overhead, self.read_totfrac,
                                self.spill_limit))

            w_memlim = self.read_overhead*st.vismem*1e9

            tot_memlim = self.read_totfrac*sum([v['memory_limit']
                                                for v in itervalues(self.client.scheduler_info()['workers'])
                                                if 'READER' in v['resources']])

            t0 = time.Time.now().unix
            timeout = 0.8*st.metadata.inttime*st.metadata.nints  # bit shorter than scan
            elapsedtime = time.Time.now().unix - t0

            if segments is None:
                segments = list(range(st.nsegment))

            while (elapsedtime < timeout) and (len(futures) < len(segments)):
                # Submit if workers are not overloaded
                if (heuristics.reader_memory_ok(self.client, w_memlim) and
                   heuristics.readertotal_memory_ok(self.client, tot_memlim) and
                   heuristics.spilled_memory_ok(limit=self.spill_limit,
                                                daskdir=self.daskdir)):
                    futures = pipeline.pipeline_scan_delayed(st,
                                                             segments=segments,
                                                             cl=self.client,
                                                             cfile=cfile,
                                                             vys_timeout=vys_timeout)
                else:
                    sleep(min(1, timeout/10))
                    self.cleanup()
                    elapsedtime = time.Time.now().unix - t0

            if elapsedtime > timeout:
                logger.info("Throttle timed out. ScanId {0} not submitted."
                            .format(st.metadata.scanId))

        else:
            logger.info('Starting pipeline...')
            futures = pipeline.pipeline_scan_delayed(st, segments=segments,
                                                     cl=self.client,
                                                     cfile=cfile,
                                                     vys_timeout=vys_timeout)

        if len(futures):
            self.futures[st.metadata.scanId] = futures

    def cleanup(self, badstatuslist=['cancelled', 'error', 'lost']):
        """ Clean up job list.
        Scans futures, removes finished jobs, and pushes results to relevant indices.
        badstatuslist can include 'cancelled', 'error', 'lost'.
        """

        removed = 0
        cindexed = 0
        sdms = 0

        scanIds = [scanId for scanId in self.futures]
        if len(scanIds):
            logger.info("Checking on {0} jobs with scanId: {1}"
                        .format(len(scanIds), ','.join(scanIds)))

        for scanId in self.futures:
            # create list of futures (a dict per segment) that are done
            finishedlist = [futures for (scanId0, futurelist) in iteritems(self.futures)
                            for futures in futurelist
                            if (futures['candcollection'].status == 'finished') and
                               (scanId0 == scanId)]

            # one candcollection per segment
            for futures in finishedlist:

                # index cands
                # option 1: pull results over
                candcollection = futures['candcollection'].result()
                ncands = len(candcollection)
                # option 2: check in place (kinda requires multi-thread gpu workers)
                # if ncands.result():
                if ncands:
#                    candcollection = futures['candcollection'].result()
                    if self.indexresults:
                        res = elastic.indexcands(candcollection, scanId,
                                                 tags=self.tags,
                                                 url_prefix=_candplot_url_prefix,
                                                 indexprefix=self.indexprefix)
                        # TODO: makesumaryplot logs cands in all segments
                        # this is confusing when only one segment being handled here
                        makesummaryplot(candcollection.prefs.workdir, scanId)
                        nplots = moveplots(candcollection, scanId,
                                           destination=_candplot_dir)
                        if res or nplots:
                            logger.info('Indexed {0} cands to {1} and '
                                        'moved {2} plots to {3} for '
                                        'scanId {4}'
                                        .format(res, self.indexprefix+'cands',
                                                nplots, _candplot_dir,
                                                scanId))
                        else:
                            logger.info('No candidates or plots found.')

                        cindexed += res
                    else:
                        logger.info("Not indexing cands found in scanId {0}"
                                    .format(scanId))
                else:
                    logger.debug('No candidates for a segment from scanId {0}'
                                 .format(scanId))

                # index noises
                noisefile = self.states[scanId].noisefile
                if os.path.exists(noisefile):
                    if self.indexresults:
                        res = elastic.indexnoises(noisefile, scanId,
                                                  indexprefix=self.indexprefix)
                        if res:
                            logger.info("Indexed {0} noises to {1} for scanId "
                                        "{2}".format(res,
                                                     self.indexprefix+"noise",
                                                     scanId))
                    else:
                        logger.debug("Not indexing noises for scanId {0}."
                                     .format(scanId))
                else:
                    logger.debug('No noisefile found, no noises indexed.')

                # optionally save and archive sdm/bdfs for segment
                if self.saveproducts and ncands:
                    newsdms = createproducts(candcollection, futures['data'])
                    sdms += len(newsdms)
                    if len(newsdms):
                        logger.info("Created new SDMs at: {0}"
                                    .format(newsdms))
                    else:
                        logger.info("No new SDMs created")

                    if self.archiveproducts:
                        runingest(newsdms)  # TODO: implement this

                else:
                    logger.debug("Not making new SDMs or moving candplots.")

                # remove job from list
                self.futures[scanId].remove(futures)
                removed += 1

        # clean up bad futures
        for status in badstatuslist:
            removed += self.removefutures(status)

        # after scanId loop, clean up self.futures
        removeids = [scanId for scanId in self.futures
                     if len(self.futures[scanId]) == 0]
        for scanId in removeids:
            _ = self.futures.pop(scanId)
            _ = self.states.pop(scanId)
            logger.info("No jobs of scanId {0} left. "
                        "Cleaning state and futures dicts".format(scanId))

#        _ = self.client.run(gc.collect)
        if removed:
            logger.info('Removed {0} jobs, indexed {1} cands, made {2} SDMs.'
                        .format(removed, cindexed, sdms))

    def cleanup_loop(self, timeout=None):
        """ Clean up until all jobs gone or timeout elapses.
        """

        if timeout is None:
            timeout_string = "no "
        else:
            timeout_string = "{0} s ".format(timeout)

        logger.info("Cleaning up all futures with {0} timeout"
                    .format(timeout_string))

        t0 = time.Time.now().unix
        elapsedtime = time.Time.now().unix - t0
        badstatuslist = ['cancelled', 'error', 'lost']
        while len(self.futures):
            if elapsedtime > timeout:
                badstatuslist += ['pending']
            self.cleanup(badstatuslist=badstatuslist)
            sleep(10)

    def inject_transient(self, scanId):
        """ Randomly sets preferences for scan to injects a transient
        into one segment of a scan. Requires state to have been set.
        Transient should be detectable, but randomly placed in the scan.
        Also pushes mock properties to index.
        """

        assert scanId in self.states, "State must be defined first"

        if random.uniform(0, 1) < self.mockprob:
            st = self.states[scanId]
            self.states[scanId].prefs.simulated_transient = make_transient(st)
            mindexed = (elastic.indexmocks(self.states[scanId],
                                           indexprefix=self.indexprefix)
                        if self.indexresults else 0)
            if mindexed:
                logger.info("Indexed {0} mock transient to {1}."
                            .format(mindexed, self.indexprefix+"mocks"))
            else:
                logger.info("Not indexing mock transient.")

    def removefutures(self, status, keep=False):
        """ Remove jobs with status of 'cancelled'
        keep argument defines whether to remove or move to futures_removed
        """

        removed = 0
        for scanId in self.futures:

            # create list of futures (a dict per segment) that are cancelled
            removelist = [futures for (scanId0, futurelist) in iteritems(self.futures)
                          for futures in futurelist
                          if (futures['candcollection'].status == status or
                              futures['data'].status == status) and
                             (scanId0 == scanId)]

            # clean them up
            for futures in removelist:
                self.futures[scanId].remove(futures)
                removed += 1

                if keep:
                    if scanId not in self.futures_removed:
                        self.futures_removed[scanId] = []
                    self.futures_removed[scanId].append(futures)

        if removed:
            logger.warn("{0} bad jobs removed from scanId {1}".format(removed,
                                                                      scanId))

        return removed

    def handle_finish(self, dataset):
        """ Triggered when obs doc defines end of a script.
        """

        logger.info('End of scheduling block message received.')


def make_transient(st):
    """ Given a state, create a randomized transient that is detectable.
    """

    segment = random.choice(range(st.nsegment))
    dmind = random.choice(range(len(st.dmarr)))
    i0 = random.choice(st.get_search_ints(segment, dmind, 0))
    dm = st.dmarr[dmind]  # TODO: allow off center DM
    dt = random.choice(st.dtarr)*st.metadata.inttime
    amp = 0.1  # TODO: make this work for sim and real data
    if random.choice([0, 1]):  # flip a coin to set either l or m
        l = math.radians(random.uniform(-st.fieldsize_deg/2, st.fieldsize_deg/2))
        m = 0.
    else:
        l = 0.
        m = math.radians(random.uniform(-st.fieldsize_deg/2, st.fieldsize_deg/2))

    return [(segment, i0, dm, dt, amp, l, m)]


def search_config(config, preffile=None, inprefs={}, nameincludes=None,
                  searchintents=None):
    """ Test whether configuration specifies a scan config that realfast should
    search
    """

    # find config properties of interest
    intent = config.scan_intent
    antennas = config.get_antennas()
    antnames = [str(ant.name) for ant in antennas]
    subbands = config.get_subbands()
    inttimes = [subband.hw_time_res for subband in subbands]
    pols = [subband.pp for subband in subbands]
    nchans = [subband.spectralChannels for subband in subbands]
    chansizes = [subband.bw/subband.spectralChannels for subband in subbands]
    reffreqs = [subband.sky_center_freq*1e6 for subband in subbands]

    # Do not process if...
    # 1) chansize changes between subbands
    if not all([chansizes[0] == chansize for chansize in chansizes]):
        logger.warn("Channel size changes between subbands: {0}"
                    .format(chansizes))
        return False

    # 2) start and stop time is after current time
    now = time.Time.now().unix
    startTime = time.Time(config.startTime, format='mjd').unix
    stopTime = time.Time(config.stopTime, format='mjd').unix
    if (startTime < now) and (stopTime < now):
        logger.warn("Scan startTime and stopTime are in the past ({0}, {1} < {2})"
                    .format(startTime, stopTime, now))
        return False

    # 3) if nameincludes set, reject if datasetId does not have it
    if nameincludes is not None:
        if nameincludes not in config.datasetId:
            logger.warn("datasetId {0} does not include nameincludes {1}"
                        .format(config.datasetId, nameincludes))
            return False

    # 4) only search if in searchintents
    if searchintents is not None:
        if not any([searchintent in intent for searchintent in searchintents]):
            logger.warn("intent {0} not in searchintents list {1}"
                        .format(intent, searchintents))
            return False

    # 5) only two antennas
    if len(antnames) <= 2:
        logger.warn("Only {0} antennas in array".format(len(antnames)))
        return False

    # 6) only if state validates
    if not heuristics.state_validates(config=config, preffile=preffile,
                                      inprefs=inprefs):
        logger.warn("State not valid for scanId {0}"
                    .format(config.scanId))
        return False

    return True


def summarize(config):
    """ Print summary info for config
    """

    try:
        logger.info(':: ConfigID {0} ::'.format(config.configId))
        logger.info('\tScan {0}, source {1}, intent {2}'
                    .format(config.scanNo, config.source,
                            config.scan_intent))

        logger.info('\t(RA, Dec) = ({0}, {1})'
                    .format(config.ra_deg, config.dec_deg))
        subbands = config.get_subbands()
        reffreqs = [subband.sky_center_freq for subband in subbands]
        logger.info('\tFreq: {0} - {1}'
                    .format(min(reffreqs), max(reffreqs)))

        nchans = [subband.spectralChannels for subband in subbands]
        chansizes = [subband.bw/subband.spectralChannels
                     for subband in subbands]
        sb0 = subbands[0]
        logger.info('\t(nspw, chan/spw, nchan) = ({0}, {1}, {2})'
                    .format(len(nchans), nchans[0], sum(nchans)))
        logger.info('\t(BW, chansize) = ({0}, {1}) MHz'
                    .format(sb0.bw, chansizes[0]))
        if not all([chansizes[0] == chansize for chansize in chansizes]):
            logger.info('\tNot all spw have same configuration.')

        logger.info('\t(nant, npol) = ({0}, {1})'
                    .format(config.numAntenna, sb0.npp))
        dt = 24*3600*(config.stopTime-config.startTime)
        logger.info('\t(StartMJD, duration) = ({0}, {1}s).'
                    .format(config.startTime, round(dt, 1)))
        logger.info('\t({0}/{1}) ints at (HW/Final) integration time of ({2:.3f}/{3:.3f}) s'
                    .format(int(round(dt/sb0.hw_time_res)),
                            int(round(dt/sb0.final_time_res)),
                            sb0.hw_time_res, sb0.final_time_res))
    except:
        logger.warn("Failed to fully parse config to print summary."
                    "Proceeding.")


def createproducts(candcollection, datafuture, sdmdir='.',
                   savebdfdir='/lustre/evla/wcbe/data/no_archive/'):
    """ Create SDMs and BDFs for a given candcollection (time segment).
    Takes data future and calls data only if windows found to cut.
    This uses the sdm_builder module, which calls the sdm builder server.
    sdmdir will move and rename output file to "realfast_<obsid>_<uid>".
    Currently BDFs are moved to no_archive lustre area by default.
    """

    if len(candcollection.array) == 0:
        logger.info('No candidates to generate products for.')
        return []

    metadata = candcollection.metadata
    segment = candcollection.segment
    if not isinstance(segment, int):
        logger.warn("Cannot get unique segment from candcollection ({0})".format(segment))
    st = candcollection.state

    candranges = gencandranges(candcollection)  # finds time windows to save from segment
    logger.info('Getting data for candidate time ranges {0}.'.format(candranges))

    data = datafuture.result()
    ninttot, nbl, nchantot, npol = data.shape
    nchan = metadata.nchan_orig//metadata.nspw_orig
    nspw = metadata.nspw_orig

    sdmlocs = []
    for (startTime, endTime) in candranges:
        i = (86400*(startTime-st.segmenttimes[segment][0])/metadata.inttime).astype(int)
        nint = (86400*(endTime-startTime)/metadata.inttime).astype(int)  # TODO: may be off by 1
        logger.info("Cutting {0} ints from int {1} for candidate at {2}"
                    .format(nint, i, startTime))
        data_cut = data[i:i+nint].reshape(nint, nbl, nspw, 1, nchan, npol)

        sdmloc = sdm_builder.makesdm(startTime, endTime, metadata, data_cut)
        if sdmloc is not None:
            if sdmdir is not None:
                uid = ('uid:///evla/realfastbdf/{0}'
                       .format(int(time.Time(startTime,
                                             format='mjd').unix*1e3)))
                if 'sdm-builder' in sdmloc:  # remove default name suffix
                    sdmlocbase = '-'.join(sdmloc.rsplit('-')[:-3])
                else:
                    sdmlocbase = sdmloc
                newsdmloc = os.path.join(sdmdir, 'realfast_{0}_{1}'
                                         .format(os.path.basename(sdmlocbase),
                                                 uid.rsplit('/')[-1]))
                assert not os.path.exists(newsdmloc), "newsdmloc {0} already exists".format(newsdmloc)
                shutil.move(sdmloc, newsdmloc)
                sdmlocs.append(newsdmloc)
            else:
                sdmlocs.append(sdmloc)

            # TODO: migrate bdfdir to newsdmloc once ingest tool is ready
            sdm_builder.makebdf(startTime, endTime, metadata, data_cut,
                                bdfdir=savebdfdir)
        else:
            logger.warn("No sdm/bdf made for start/end time {0}-{1}"
                        .format(startTime, endTime))

    return sdmlocs


def gencandranges(candcollection):
    """ Given a candcollection, define a list of candidate time ranges.
    """

    segment = candcollection.segment
    st = candcollection.state

    # save whole segment
    return [(st.segmenttimes[segment][0], st.segmenttimes[segment][1])]


def runingest(sdms):
    """ Call archive tool or move data to trigger archiving of sdms.
    This function will ultimately be triggered by candidate portal.
    """

    NotImplementedError
#    /users/vlapipe/workflows/test/bin/ingest -m -p /home/mctest/evla/mcaf/workspace --file 


def moveplots(candcollection, scanId, destination=_candplot_dir):
    """ For given fileroot, move candidate plots to public location
    """

    workdir = candcollection.prefs.workdir
    segment = candcollection.segment

    logger.info("Moving plots for scanId {0} segment {1} to {2}"
                .format(scanId, segment, destination))

    nplots = 0
    candplots = glob.glob('{0}/cands_{1}_seg{2}-*.png'
                          .format(workdir, scanId, segment))
    for candplot in candplots:
        try:
            shutil.move(candplot, destination)
            nplots += 1
        except shutil.Error:
            logger.warn("Plot {0} already exists at {1}. Skipping..."
                        .format(candplot, destination))

    # move summary plot too
    summaryplot = '{0}/cands_{1}.html'.format(workdir, scanId)
    summaryplotdest = os.path.join(destination, os.path.basename(summaryplot))
    if os.path.exists(summaryplot):
        shutil.move(summaryplot, summaryplotdest)
    else:
        logger.warn("No summary plot {0} found".format(summaryplot))

    return nplots


def makesummaryplot(workdir, scanId):
    """ Create summary plot for a given scanId and move it
    """

    candsfile = '{0}/cands_{1}.pkl'.format(workdir, scanId)
    candidates.makesummaryplot(candsfile)


class config_controller(Controller):

    def __init__(self, pklfile=None, preffile=_preffile):
        """ Creates controller object that saves scan configs.
        If pklfile is defined, it will save pickle there.
        If preffile is defined, it will attach a preferences to indexed scan.
        Inherits a "run" method that starts asynchronous operation.
        """

        super(config_controller, self).__init__()
        self.pklfile = pklfile
        self.preffile = preffile

    def handle_config(self, config):
        """ Triggered when obs comes in.
        Downstream logic starts here.
        """

        logger.info('Received complete configuration for {0}, '
                    'scan {1}, source {2}, intent {3}'
                    .format(config.scanId, config.scanNo, config.source,
                            config.scan_intent))

        if self.pklfile:
            with open(self.pklfile, 'ab') as pkl:
                pickle.dump(config, pkl)

        if self.preffile:
            prefs = preferences.Preferences(**preferences.parsepreffile(self.preffile,
                                                                        name='default'))
            elastic.indexscan_config(config, preferences=prefs)



from setuptools import setup, find_packages
import glob

setup(
    name='realfast',
    description='Real-time data analysis at the VLA',
    author='Casey Law and the realfast team',
    author_email='caseyjlaw@gmail.com',
    version='3.3.1',
    url='http://realfast.io',
    include_package_data=True,
    packages=find_packages(),
    package_data={'realfast': ['xsd/*.xsd']},
    data_files=[('conf', glob.glob('conf/*'))],
#    scripts=['scripts/sdm_chop-n-serve.pl'],   # add non-python scripts
    install_requires=['rfpipe', 'evla_mcast', 'sdmpy', 'click',
                      'elasticsearch', 'distributed', 'future'],
                      # 'rfgpu', 'vysmaw_reader'],
    entry_points='''
        [console_scripts]
        realfast=realfast.cli:cli
'''
)



from __future__ import print_function, division, absolute_import #, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

from realfast import controllers
import click

import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logging.captureWarnings(True)
logger = logging.getLogger(__name__)

vys_cfile = '/home/cbe-master/realfast/soft/vysmaw_apps/vys.conf'
default_preffile = '/lustre/evla/test/realfast/realfast.yml'
default_vys_timeout = 10  # seconds more than segment length
distributed_host = 'cbe-node-01'


@click.group('realfast')
def cli():
    pass


@cli.command()
@click.option('--pklfile')
@click.option('--preffile')
def config_catcher(pklfile, preffile):
    """ Runs async process to catch multicast messages to form scan config objects
    Can be saved to pklfile and optionally attached to preferences from preffile.
    """

    logger.info("Running config catcher with pklfile={0} and preffile={1}"
                .format(pklfile, preffile))
    config = controllers.config_controller(pklfile=pklfile, preffile=preffile)
    config.run()


@cli.command()
@click.option('--preffile', default=default_preffile)
@click.option('--vys_timeout', default=default_vys_timeout)
def run(preffile, vys_timeout):
    """ Run realfast controller to catch scan configs and start rfpipe.
    """

    logger.warn("Command line realfast tool will lose track of jobs and do " 
                "no indexing after ctrl-c.")
    rfc = controllers.realfast_controller(preffile=preffile,
                                          vys_timeout=vys_timeout)
    rfc.run()
